/*
 *  Copyright (C) 2014 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 * This code is proprietary code subject to the terms and conditions 
 * defined in LICENSE file which is part of this source code package.
 *
 */

#if 1
#define PRINT_TRACE() {}
#else
#define PRINT_TRACE()                               \
  {                                                 \
    traceEvent(TRACE_NORMAL,			    \
    "%s (%s:%d)\n",                                 \
	    __FUNCTION__, __FILE__, __LINE__);      \
    }
#endif

/*
  http://emailcommand.wordpress.com/linux/how-does-inode-in-procnettcp-related-with-the-procpidfdnum-socketnum/

  STEP TO KNOW WHICH SOCKET IS OWN BY APPLICATION PROGRAM IN LINUX OPERATING SYSTEM.

  1. For TCP socket scans /proc/net/tcp to obtain the inode number.
  2. Scans /proc/[1-9]<star>/fd/<star> for a reference to the inode.
  3. If you know the process id (PID) ,
     ( the relevant inode in /proc/net/tcp to the reference inode in proc/PID/fd/ )
  4. The application which generate the process PID can be list out as
     $readlink /proc/PID/exe
*/

#define MAX_PORT         65536

/* ******************************************* */

void read_proto_sockets(char *path, u_int8_t is_tcp, u_int8_t ip_version) {
  FILE *fd = fopen(path, "r");
  char buf[256], *line;

  if(fd == NULL) return;

  line = fgets(buf, sizeof(buf), fd); /* Discard header */

  while((line = fgets(buf, sizeof(buf), fd)) != NULL) {
    u_int16_t _localport, _remoteport;
    u_int64_t _inode;
    char      *_localhost;

    /*
      sl  local_address rem_address   st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
      0: 0100007F:15B4 00000000:0000 0A 00000000:00000000 00:00000000 00000000  1000        0 27138 1 0000000000000000 100 0 0 10 0
    */
    // traceEvent(TRACE_WARNING, "%s", line);
    strtok(line, ":"); 

    strtok(NULL, ":");
    strtol(strtok(NULL, " "), NULL, 16);
    strtok(NULL, ":"); 
    _remoteport = strtol(strtok(NULL, " "), NULL, 16);
    strtok(NULL, " ");strtok(NULL, " ");strtok(NULL, " ");strtok(NULL, " ");strtok(NULL, " ");strtok(NULL, " ");
    _inode = atoll(strtok(NULL, " "));

    // traceEvent(TRACE_WARNING, "[%s%u] [%u:%u] [%llu]", is_tcp ? "TCP" : "UDP", ip_version, _localport, _remoteport, _inode);

    if(_inode != 0) {
      struct port_inode *p = (struct port_inode*)malloc(sizeof(struct port_inode));

      if(p) {
	p->local_port = _localport, p->remote_port = _remoteport, p->inode = _inode, p->is_tcp = is_tcp, p->ip_version = ip_version;
	HASH_ADD_INT64(readWriteGlobals->sysdig.currently_open_ports, inode, p);
      }

      // traceEvent(TRACE_WARNING, "[%s%u] [%u:%u] [%llu]", is_tcp ? "TCP" : "UDP", ip_version, _localport, _remoteport, _inode);
    }
  }

  fclose(fd);
}

/* ******************************************* */

char* get_process_name(u_int32_t pid, char *buf, u_int buf_len) {
  if(pid > 0) {
    char path[256];
    FILE *f;

    buf[0] = '\0';

    snprintf(path, sizeof(path), "/proc/%u/status", pid);

    if((f = fopen(path, "r")) != NULL) {
      char *line;

      while((line = fgets(buf, buf_len, f)) != NULL) {
	if(strncmp(line, "Name:", 5) == 0) {
	  int l = strlen(buf)-1;
	  if(buf[l] == '\n') buf[l] = '\0';
	  fclose(f);
	  return(&buf[6]);
	}
      }

      fclose(f);
      return(buf);
    } else
      traceEvent(TRACE_INFO, "Unable to read %s", path);
  }

  snprintf(buf, buf_len, "???");
  return(buf);
}

/* ******************************************* */

void delete_port_from_cache(u_int8_t is_tcp, u_int8_t ip_version,
			    u_int16_t the_local_port, u_int16_t the_remote_port) {
  struct port_exe *found;
  u_int32_t pkey = (the_local_port << 16) + the_remote_port; /* 32 bit key */
  u_int32_t expire_in = readOnlyGlobals.lifetimeTimeout*2; /* sec */
  struct port_exe *pe;

  if(ip_version == 4)
    pe = is_tcp ? readWriteGlobals->sysdig.port_bound_to_exe_tcp4 : readWriteGlobals->sysdig.port_bound_to_exe_udp4;
  else
    pe = is_tcp ? readWriteGlobals->sysdig.port_bound_to_exe_tcp6 : readWriteGlobals->sysdig.port_bound_to_exe_udp6;

  pthread_rwlock_wrlock(&cacheLock);
  HASH_FIND_INT32(pe, &pkey, found);
  if(found) {
    if(0) {
      if(is_tcp) {
	if(ip_version == 4)
	  HASH_DEL(readWriteGlobals->sysdig.port_bound_to_exe_tcp4, found);
	else
	  HASH_DEL(readWriteGlobals->sysdig.port_bound_to_exe_tcp6, found);
      } else {
	if(ip_version == 4)
	  HASH_DEL(readWriteGlobals->sysdig.port_bound_to_exe_udp4, found);
	else
	  HASH_DEL(readWriteGlobals->sysdig.port_bound_to_exe_udp6, found);
      }
      if(found->exe) free(found->exe);
      free(found);
    } else {
      found->expire_time = readWriteGlobals->now + expire_in;
    }

    //    if(readOnlyGlobals.enable_debug)
    {
      PRINT_TRACE();
      traceEvent(TRACE_INFO, "[SPROBE] Deleted [%s%d][port: %u:%u][found: %u]",
		 is_tcp ? "TCP" : "UDP", ip_version,
		 the_local_port, the_remote_port, found);
    }
  }

  pthread_rwlock_unlock(&cacheLock);
}

/* ******************************************* */

u_int16_t get_father_pid(u_int16_t pid) {
  char path[256], what[256];
  FILE *f;
  u_int16_t father_pid = 0;

  snprintf(path, sizeof(path), "/proc/%u/stat", pid);
  if((f = fopen(path, "r")) != NULL) {
    char *s, *line;

    /* 7452 (sshd) S 988 7452 7452 */
    line = fgets(what, sizeof(what), f);
    s = strchr(line,' ');
    s = strchr(&s[1],' ');
    s = strchr(&s[1],' ');

    father_pid = atoi(&s[1]);
    fclose(f);
  }

  return(father_pid);
}

/* ******************************************* */

/*
            $ cat /proc/$$/status
                  Name:   bash
                  State:  S (sleeping)
                  Tgid:   3515
                  Pid:    3515
                  PPid:   3452
                  TracerPid:      0
                  Uid:    1000    1000    1000    1000
                  Gid:    100     100     100     100
*/

void get_uid_gid(u_int16_t pid, u_int16_t *uid, u_int16_t *gid) {
  if(pid > 0) {
    char path[256], buf[256];
    FILE *f;

    snprintf(path, sizeof(path), "/proc/%u/status", pid);

    if((f = fopen(path, "r")) != NULL) {
      char *line;

      while((line = fgets(buf, sizeof(buf), f)) != NULL) {
	if(strncmp(line, "Uid:", 4) == 0)
	  *uid = atoi(&line[5]);
	else if(strncmp(line, "Gid:", 4) == 0) {
	  *gid = atoi(&line[5]);
	  break;
	}
      }

      fclose(f);
    } else
      traceEvent(TRACE_INFO, "Unable to read %s", path);
  }
}

/* ******************************************* */

void set_port_into_cache(u_int8_t client_port, u_int8_t is_tcp,
			 u_int8_t ip_version,
			 u_int16_t the_local_port,
			 u_int16_t the_remote_port,
			 u_int16_t pid, char *process_name) {
  struct port_exe *p, *found;
  u_int32_t pkey = (the_local_port << 16) + the_remote_port; /* 32 bit key */
  struct port_exe *pe;
  u_int8_t added = 0;

  if(ip_version == 4)
    pe = is_tcp ? readWriteGlobals->sysdig.port_bound_to_exe_tcp4 : readWriteGlobals->sysdig.port_bound_to_exe_udp4;
  else
    pe = is_tcp ? readWriteGlobals->sysdig.port_bound_to_exe_tcp6 : readWriteGlobals->sysdig.port_bound_to_exe_udp6;

  pthread_rwlock_wrlock(&cacheLock);
  HASH_FIND_INT32(pe, &pkey, found);
  if(found) {
    /* Recycle entry already existing */
    if(found->exe) free(found->exe);
    found->exe = strdup(process_name);

    if(found->pid != pid) {
      found->pid = pid;

      found->father_pid = get_father_pid(pid);
      added = 1;
    }
    found->expire_time = 0;
  } else {
    p = (struct port_exe*)malloc(sizeof(struct port_exe));

    if(p) {
      p->port = pkey, p->pid = pid, p->expire_time = 0,
	p->father_pid = get_father_pid(pid), p->exe = strdup(process_name);
      if(is_tcp) {
	if(ip_version == 4)
	  HASH_ADD_INT32(readWriteGlobals->sysdig.port_bound_to_exe_tcp4, port, p);
	else
	  HASH_ADD_INT32(readWriteGlobals->sysdig.port_bound_to_exe_tcp6, port, p);
      } else {
	if(ip_version == 4)
	  HASH_ADD_INT32(readWriteGlobals->sysdig.port_bound_to_exe_udp4, port, p);
	else
	  HASH_ADD_INT32(readWriteGlobals->sysdig.port_bound_to_exe_udp6, port, p);
      }
    }
  }

  pthread_rwlock_unlock(&cacheLock);

  // if(readOnlyGlobals.enable_debug)
  {
    PRINT_TRACE();
    traceEvent(TRACE_INFO, "[SPROBE] Added [%s%d][port: %u:%u][%u/%s]",
	       is_tcp ? "TCP" : "UDP", ip_version, the_local_port, the_remote_port, pid, process_name);
  }
}

// FIX - Add periodic purge of sysdig.port_bound_to_exe for entries whose expire_time > now

/* ******************************************* */

void add_tid_pid_exe(u_int16_t tid, u_int16_t pid, u_int16_t father_pid,
		     u_int16_t uid, u_int16_t gid, char *_exe) {
  struct tid_pid_exe *found = NULL;
  u_int32_t key = tid;
  char exe[128], *buddy;

  if(_exe == NULL)
    snprintf(exe, sizeof(exe), "???");
  else {
    char *space, *slash = strrchr(_exe, '/');

#if 0
    if(strstr(_exe, "lt-nprobe"))
      traceEvent(TRACE_NORMAL, "add_tid_pid_exe(pid=%u/%u, %s)", pid, tid, _exe);
#endif

    if(slash)
      _exe = &slash[1];

    if(_exe[0] == '-') _exe++;

    snprintf(exe, sizeof(exe), "%s", _exe);
    space = strrchr(exe, ' ');
    if(space) space[0] = '\0';
  }

  if((uid == 0) && (gid == 0))
    get_uid_gid(pid, &uid, &gid);

  HASH_FIND_INT32(readWriteGlobals->sysdig.tid2pid, &key, found);
  if(found) {
    found->pid = pid, found->father_pid = father_pid, found->uid = uid, found->gid = gid, found->expire_time = 0;
    if(found->exe) free(found->exe);
    _exe = found->exe = strdup(exe);

    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Updating mapping [tid=%u][pid=%u][exe=%s]", tid, pid, exe);
  } else {
    struct tid_pid_exe *p = (struct tid_pid_exe*)malloc(sizeof(struct tid_pid_exe));

    if(p) {
      p->tid = tid, p->pid = pid, p->father_pid = father_pid, p->expire_time = 0,
	p->uid = uid, p->gid = gid, _exe = p->exe = strdup(exe);
      HASH_ADD_INT32(readWriteGlobals->sysdig.tid2pid, tid, p);

      if(readOnlyGlobals.enable_debug)
	traceEvent(TRACE_NORMAL, "#### Added mapping [tid=%u][pid=%u][exe=%s]", tid, pid, exe);
    } else
      return;
  }

  /*
    We do this at the end so that we do not modify the original _exe
    that can lead to a crash
  */
  if((buddy = strchr(_exe, ':')) != NULL) buddy[0] = '\0';
  if((buddy = strchr(_exe, ' ')) != NULL) buddy[0] = '\0';
}

/* ******************************************* */

void delete_tid(u_int16_t tid) {
  struct tid_pid_exe *found = NULL;
  u_int32_t key = tid; /* it must be 32 bit as it's the key */
  char what[256];
  struct stat buf;

  snprintf(what, sizeof(what), "/proc/%u", tid);
  if(stat(what, &buf) == 0) {
    /* Process is still active (perhaps it's a server): discarded */
    return;
  }

  HASH_FIND_INT32(readWriteGlobals->sysdig.tid2pid, &key, found);

  if(found) {
    if(0) {
      HASH_DEL(readWriteGlobals->sysdig.tid2pid, found);
      if(found->exe) free(found->exe);
      free(found);
    } else {
      u_int32_t expire_in = readOnlyGlobals.lifetimeTimeout*2; /* sec */

      found->expire_time = readWriteGlobals->now + expire_in;
    }

    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Deleted mapping [tid=%u]", tid);
  } else {
    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Unable to delete mapping [tid=%u]", tid);
  }
}

// FIX - Add periodic purge of sysdig.tid2pid for entries whose expire_time > now

/* ******************************************* */

u_int16_t tid_to_pid(u_int16_t tid, char **exe) {
  struct tid_pid_exe *found = NULL;
  u_int32_t key = tid; /* it must be 32 bit as it's the key */

  HASH_FIND_INT32(readWriteGlobals->sysdig.tid2pid, &key, found);
  if(found) {
    *exe = found->exe;
    return(found->pid);
  } else {
    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Unable to find mapping [tid=%u]", tid);
  }

  *exe = "???";
  return(tid);
}

/* ******************************************* */

void read_known_users() {
  struct passwd *p;

  while ((p = getpwent()) != NULL) {
    struct uid_name *uid = (struct uid_name*)malloc(sizeof(struct uid_name));

    if(!uid) break;

    uid->uid = p->pw_uid, uid->gid = p->pw_gid, uid->user_name = strdup(p->pw_name);
    HASH_ADD_INT32(readWriteGlobals->sysdig.uid_to_name, uid, uid);
  }

  endpwent();
}

/* ******************************************* */

void read_open_sockets() {
  DIR *dirp, *sub_dirp;
  struct dirent *dp, *sub_dp;
  struct port_inode *current, *tmp;
  int id = readWriteGlobals->now % MAX_NUM_REDIS_CONNECTIONS;

  traceEvent(TRACE_INFO, "Reading socket state");
  read_proto_sockets("/proc/net/tcp", 1, 4);
  read_proto_sockets("/proc/net/tcp6", 1, 6);
  read_proto_sockets("/proc/net/udp", 0, 4);
  read_proto_sockets("/proc/net/udp6", 0, 6);

  if((dirp = opendir("/proc")) != NULL) {
    while((dp = readdir(dirp)) != NULL) {
      if(dp->d_name && (isdigit(dp->d_name[0]))) {
	char path[256], process_name[128], *name;
	long pid = atol(dp->d_name);

	name = get_process_name(pid, process_name, sizeof(process_name));

	/* Read threads */
	snprintf(path, sizeof(path), "/proc/%s/task", dp->d_name);

	if((sub_dirp = opendir(path)) != NULL) {
	  while((sub_dp = readdir(sub_dirp)) != NULL) {
	    if(sub_dp->d_name && (sub_dp->d_name[0] != '.')) {
	      int rc;
	      char what[256], sym[256];

	      snprintf(what, sizeof(what), "%s/%s/exe", path, sub_dp->d_name);

	      rc = readlink(what, sym, sizeof(sym));

	      if(rc > 0) {
		u_int16_t pid = atoi(dp->d_name);
		u_int16_t father_pid = get_father_pid(pid);

		sym[rc] = 0;

		if(readOnlyGlobals.enable_debug)
		  traceEvent(TRACE_INFO, "[tid=%s][pid=%u][father_pid=%u][exe=%s]",
			     sub_dp->d_name, pid, father_pid, sym);
		add_tid_pid_exe(atoi(sub_dp->d_name), pid, father_pid, 0 /* uid */, 0 /* gid */, sym);
	      }
	    }
	  }
	}

	/* Read FDs */
	snprintf(path, sizeof(path), "/proc/%s/fd", dp->d_name);

	if((sub_dirp = opendir(path)) != NULL) {
	  while((sub_dp = readdir(sub_dirp)) != NULL) {
	    if(sub_dp->d_name && (sub_dp->d_name[0] != '.')) {
	      struct stat buf;
	      char what[256];

	      snprintf(what, sizeof(what), "%s/%s", path, sub_dp->d_name);

	      if((stat(what, &buf) == 0) && S_ISSOCK(buf.st_mode)) {
		struct port_inode *p;
		u_int64_t inode = buf.st_ino;

		HASH_FIND_INT64(readWriteGlobals->sysdig.currently_open_ports, &inode, p);

		if(p)
		  set_port_into_cache(0, p->is_tcp, p->ip_version, p->local_port, p->remote_port, pid, name);
	      }
	    }
	  }
	}
      }
    }

    closedir(dirp);
  }

  HASH_ITER(hh, readWriteGlobals->sysdig.currently_open_ports, current, tmp) {
    HASH_DEL(readWriteGlobals->sysdig.currently_open_ports, current);
    free(current);
  }
}

/* ******************************************* */

void add_socket(u_int16_t pid, int fd, u_int socket_type) {
  u_int32_t key;
  struct fd_type *found = NULL;

  if(pid == 0)
    return;

  key = (pid << 16) + fd;

  HASH_FIND_INT32(readWriteGlobals->sysdig.fd2type, &key, found);
  if(found) {
    found->socket_type = socket_type;
  } else {
    struct fd_type *pp = (struct fd_type*)malloc(sizeof(struct fd_type));

    if(pp) {
      pp->fd_pid = key, pp->socket_type = socket_type;

      HASH_ADD_INT32(readWriteGlobals->sysdig.fd2type, fd_pid, pp);
      if(readOnlyGlobals.enable_debug)
	traceEvent(TRACE_NORMAL, "#### Adding socket [fd=%u][pid=%u][socket_type=%u]",
		   fd, pid, socket_type);
    }
  }
}

/* ******************************************* */

void delete_socket(u_int16_t pid, int fd) {
  u_int32_t key;
  struct fd_type *found = NULL;

  if(pid == 0)
    return;

  key = (pid << 16) + fd;

  HASH_FIND_INT32(readWriteGlobals->sysdig.fd2type, &key, found);
  if(found) {
    HASH_DEL(readWriteGlobals->sysdig.fd2type, found);
    free(found);
  }
}

/* ******************************************* */

u_int get_socket_type(u_int16_t pid, int fd) {
  u_int32_t key;
  struct fd_type *found = NULL;

  if(pid == 0)
    return((u_int)-1);

  key = (pid << 16) + fd;

  HASH_FIND_INT32(readWriteGlobals->sysdig.fd2type, &key, found);
  if(found)
    return(found->socket_type);

  return((u_int)-1); /* Default */
}

/* ******************************************* */

void update_local_port(u_int8_t client_port, u_int8_t is_tcp,
		       u_int8_t ip_version,
		       u_int16_t the_local_port, u_int16_t the_remote_port, 
		       u_int16_t tid, int fd) {
  u_int16_t pid;
  char *process_name;
  u_int32_t key;
  struct fd_pid *found = NULL;

  if(tid == 0)
    return;

  pid = tid_to_pid(tid, &process_name);
  key = (pid << 16) + fd;

  if(0)
  traceEvent(TRACE_NORMAL, "[REDIS] [%s][port: %u/%u][%u/%s]",
	     is_tcp ? "TCP" : "UDP", the_local_port, the_remote_port, pid, process_name);

  HASH_FIND_INT32(readWriteGlobals->sysdig.active_fd_pid, &key, found);
  if(found) {
    /* This association already exists */

    if(found->local_port != the_local_port) {
      if(readOnlyGlobals.enable_debug)
	traceEvent(TRACE_NORMAL, "#### Updating port [port=%u:%u][port=%u:%u][exe=%s]",
		   found->local_port, found->remote_port, the_local_port, the_remote_port, process_name);

      if(found->process_name) free(found->process_name);

      found->process_name = strdup(process_name), found->tid = tid, found->is_tcp = is_tcp,
	found->local_port = the_local_port, found->remote_port = the_remote_port,
	found->ip_version = ip_version;
    } else
      return;
  } else {
    struct fd_pid *pp = (struct fd_pid*)malloc(sizeof(struct fd_pid));

    if(pp) {
      pp->fd_pid = key, pp->local_port = the_local_port, pp->remote_port = the_remote_port, 
	pp->when = get_current_time(),
	pp->tid = tid, process_name = pp->process_name = strdup(process_name),
	pp->is_tcp = is_tcp, pp->ip_version = ip_version;
      HASH_ADD_INT32(readWriteGlobals->sysdig.active_fd_pid, fd_pid, pp);
      if(readOnlyGlobals.enable_debug)
	traceEvent(TRACE_NORMAL, "#### Adding port %u [fd=%u][tid=%u][pid=%u][exe=%s]",
		   the_local_port, fd, tid, pid, process_name);
    }
  }

  PRINT_TRACE();
  set_port_into_cache(client_port, is_tcp, ip_version,
		      the_local_port, the_remote_port,
		      pid, process_name);
}

/* ******************************************* */

void print_tuple(u_int8_t as_client, u_int16_t *the_local_port, 
		 u_int16_t *the_remote_port,
		 char *process_name, u_int process_name_len,
		 struct sysdig_event_header* ev, int fd,
		 const struct sysdig_event_info *i, char *tuple,
		 char *proto, u_int8_t half_tuple) {
  u_int8_t m_type = tuple[0];
    u_int16_t m_sport = 0, m_dport = 0;

  *the_local_port = 0, *the_remote_port;

  if(m_type == AF_INET) {
    // u_int32_t m_sip = *((u_int32_t*)(&tuple[1]));
    // u_int32_t m_dip = *((u_int32_t*)(&tuple[1+6]));

    m_sport  = *((u_int16_t*)(&tuple[1+4]));
    m_dport  = *((u_int16_t*)(&tuple[1+10]));
  } else if(m_type == AF_INET6) {
    m_sport = *((u_int16_t*)(&tuple[1+16]));
    m_dport = *((u_int16_t*)(&tuple[1+34]));
  }

  if(as_client)
    *the_local_port = m_sport, *the_remote_port = m_dport;
  else
    *the_local_port = m_dport, *the_remote_port = m_sport;
}

/* ******************************************* */

static void create_flow(struct sysdig_event_header* ev, const struct sysdig_event_info *i,
			u_int8_t as_client, int fd, char *tuple, u_int8_t proto,
			u_int8_t ip_version) {
  char process_name[128];
  u_int16_t the_local_port, the_remote_port;

  print_tuple(as_client, &the_local_port, &the_remote_port,
	      process_name, sizeof(process_name),
	      ev, fd, i, tuple,
	      (proto == IPPROTO_TCP) ? "TCP" : "UDP", 0);

  if(the_local_port == 0) return;

  PRINT_TRACE();
  update_local_port(as_client, (proto == IPPROTO_TCP) ?  1 : 0,
		    ip_version, the_local_port, the_remote_port,
		    ev->thread_id, fd);
}

/* ******************************************* */

static void end_flow(struct sysdig_event_header* ev,
		     const struct sysdig_event_info *i, int fd) {
  struct fd_pid pp, *found = NULL;
  char *exe;
  u_int16_t pid = tid_to_pid(ev->thread_id, &exe);

  delete_socket(pid, fd);

  pp.fd_pid = (pid << 16) + fd;

  if(readOnlyGlobals.enable_debug)
    traceEvent(TRACE_INFO, "End (flow?) [%lu][fd=%d][tid=%lu][pid=%u]",
	       ev->ts, fd, ev->thread_id, pid);

  HASH_FIND_INT32(readWriteGlobals->sysdig.active_fd_pid, &pp.fd_pid, found);
  if(found) {
    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Found port to delete %u:%u", found->local_port, found->remote_port);

    /* This association already exists */
    PRINT_TRACE();
    delete_port_from_cache(found->is_tcp, found->ip_version, found->local_port, found->remote_port);
    HASH_DEL(readWriteGlobals->sysdig.active_fd_pid, found);
    free(found);
  } else {
    if(readOnlyGlobals.enable_debug)
      traceEvent(TRACE_NORMAL, "#### Unable to find port for [tid=%lu][pid=%u][fd=%d]",
		 ev->thread_id, pid, fd);
  }
}

/* ******************************************* */

void handle_event(struct sysdig_event_header* ev, u_int16_t cpuid) {
  const struct sysdig_event_info *i =  sysdig_event2info(ev->event_type);
  u_int16_t tid = ev->thread_id;

  if(0) {
    long t = ev->ts / 1000000000, now = (u_long)get_current_time();

    traceEvent(TRACE_INFO, "%lu)\t[tid=%lu][len=%u][type=%u][diff=%ld]\n",
	       ev->ts, ev->thread_id, ev->event_len, ev->event_type, now-t);
    return;
  }

  switch(ev->event_type) {
  case SYSDIG_SYSCALL_EXECVE_8_E:
  case SYSDIG_SYSCALL_EXECVE_13_E:
  case SYSDIG_CLONE_11_E:
  case SYSDIG_CLONE_16_E:
    /* Nothing to do */
    break;

  case SYSDIG_SYSCALL_EXECVE_8_X:
  case SYSDIG_SYSCALL_EXECVE_13_X:
  case SYSDIG_CLONE_11_X:
  case SYSDIG_CLONE_16_X:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      // u_int32_t res = *((u_int32_t*)valptr);
      char *exe  = (char*)(valptr+lens[0]);
      char *args = (char*)(valptr+lens[0]+lens[1]);
      u_int32_t tid = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]));
      u_int32_t pid = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]));
      u_int32_t father_pid = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]));
      char *cwd = (char*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]);
      // u_int32_t fdlimit = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]));
      // u_int32_t pgft_maj = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]));
      // u_int32_t pgft_min = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]));
      // u_int32_t vm_size = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]));
      // u_int32_t vm_rss = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]+lens[10]));
      // u_int32_t vm_swap = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]+lens[10]+lens[11]));
      u_int32_t flags = 0, uid = 0, gid = 0;

      if(ev->event_type != SYSDIG_SYSCALL_EXECVE_8_X) {
	// flags = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]+lens[10]+lens[11]+lens[12]));
	uid = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]+lens[10]+lens[11]+lens[12]+lens[13]));
	gid = *((u_int32_t*)(valptr+lens[0]+lens[1]+lens[2]+lens[3]+lens[4]+lens[5]+lens[6]+lens[7]+lens[8]+lens[9]+lens[10]+lens[11]+lens[12]+lens[13]+lens[14]));
      }

      if(readOnlyGlobals.enable_debug)
	traceEvent(TRACE_INFO, "[Event type: %u][exe: %s]", ev->event_type, exe);

      add_tid_pid_exe(tid, pid, father_pid, uid, gid, exe);
    }
    break;

  case SYSDIG_PROCEXIT_E: /* The main process ends */
    /* A process/thread is over */
    delete_tid(tid);
    break;

  case SYSDIG_SOCKET_SOCKET_E:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);

      readWriteGlobals->sysdig.last_socket_domain[cpuid] = *((u_int32_t*)(valptr));
      readWriteGlobals->sysdig.last_socket_type[cpuid]   = *((u_int32_t*)(valptr+lens[0]));
      readWriteGlobals->sysdig.last_socket_proto[cpuid]  = *((u_int32_t*)(valptr+lens[0]+lens[1]));
    }
    break;

  case SYSDIG_SOCKET_SOCKET_X:
    if((readWriteGlobals->sysdig.last_socket_domain[cpuid] == AF_INET)
       || (readWriteGlobals->sysdig.last_socket_domain[cpuid] == AF_INET6)) {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      u_int32_t fd = *((u_int32_t*)valptr);
      u_int16_t pid;
      char *process_name;

      pid = tid_to_pid(tid, &process_name);

#if 0
      if(process_name && (strcmp(process_name, "lt-nprobe") != 0))
	traceEvent(TRACE_NORMAL, "socket(fd: %u)(%s)", fd, process_name);
#endif

      add_socket(pid, fd, readWriteGlobals->sysdig.last_socket_type[cpuid]);
    }
    break;

  case SYSDIG_SOCKET_SENDTO_E:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      u_int32_t fd = *((u_int32_t*)valptr);
      // u_int32_t size = *((u_int32_t*)(valptr+lens[0]));
      char *tuple  = (char*)((valptr+lens[0]+lens[1]));
      u_int8_t ip_version = tuple[0];
      u_int sock_type;
      u_int16_t pid;
      char *process_name;

      if((ip_version != AF_INET) && (ip_version != AF_INET6)) return;

      readWriteGlobals->sysdig.last_fd[cpuid] = fd;
      pid = tid_to_pid(tid, &process_name);
      sock_type = get_socket_type(pid, readWriteGlobals->sysdig.last_fd[cpuid]);

      if((pid == my_tid) && (sock_type == (u_int)-1 /* Not found */)) {
	/* If is UDP the tuple is NOT NULL (Optimization) */
	u_int16_t m_sport  = *((u_int16_t*)(&tuple[1+4]));      
	
	if(m_sport == 0) sock_type = SOCK_STREAM; else sock_type = SOCK_DGRAM;

	// traceEvent(TRACE_NORMAL, "Guessed (sock_type=%u)(%s)", sock_type, process_name);
      }

      if(sock_type != (u_int)-1 /* Found */) {
	PRINT_TRACE();
	create_flow(ev, i, 1, readWriteGlobals->sysdig.last_fd[cpuid], tuple,
		    (sock_type == SOCK_STREAM) ? IPPROTO_TCP : IPPROTO_UDP,
		    (ip_version == AF_INET) ? 4 : 6);
      }
    }
    break;

  case SYSDIG_SYSCALL_CLOSE_E:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      u_int32_t fd = *((u_int32_t*)valptr);

      readWriteGlobals->sysdig.last_fd[cpuid] = fd;
      // PRINT_TRACE();
      end_flow(ev, i, readWriteGlobals->sysdig.last_fd[cpuid]);
    }
    break;

  case SYSDIG_SOCKET_CONNECT_E:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      u_int32_t fd = *((u_int32_t*)valptr);

      readWriteGlobals->sysdig.last_fd[cpuid] = fd;
    }
    break;

  case SYSDIG_SOCKET_CONNECT_X:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      int32_t res = *((int32_t*)valptr);
      char *tuple  = (char*)(valptr+lens[0]);
      u_int8_t ip_version = tuple[0];
      u_int sock_type;
      u_int16_t pid;
      char *process_name;

      if(((ip_version != AF_INET) && (ip_version != AF_INET6))
	 || ((res != 0) && (res != -115 /* EINPROGRESS */)))
	return;

      if(res == -115 /* EINPROGRESS */) sock_type = SOCK_STREAM; /* Optimization */
      else {
	pid = tid_to_pid(tid, &process_name);
	sock_type = get_socket_type(pid, readWriteGlobals->sysdig.last_fd[cpuid]);
      }

      if(readWriteGlobals->sysdig.last_fd[cpuid] == 53)
      traceEvent(TRACE_INFO, "SYSDIG_SOCKET_CONNECT_X [fd: %u][sock_type: %u]",
		 readWriteGlobals->sysdig.last_fd[cpuid], sock_type);

      if(sock_type != (u_int)-1 /* Not found */) {
	PRINT_TRACE();
	create_flow(ev, i, 1, readWriteGlobals->sysdig.last_fd[cpuid], tuple,
		    (sock_type == SOCK_STREAM) ? IPPROTO_TCP : IPPROTO_UDP,
		    (ip_version == AF_INET) ? 4 : 6);
      }
    }
    break;

  case SYSDIG_SOCKET_ACCEPT4_X:
  case SYSDIG_SOCKET_ACCEPT_X:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      int32_t fd = *((u_int32_t*)valptr);
      char *tuple  = (char*)(valptr+lens[0]);
      u_int8_t ip_version = tuple[0];

      if((ip_version != AF_INET) && (ip_version != AF_INET6)) return;

      readWriteGlobals->sysdig.last_fd[cpuid] = fd;

      PRINT_TRACE();
      create_flow(ev, i, 0, readWriteGlobals->sysdig.last_fd[cpuid], tuple,
		  IPPROTO_TCP, (ip_version == AF_INET) ? 4 : 6);
    }
    break;

  case SYSDIG_SOCKET_SENDMSG_E:
    {
      u_int32_t nparams = i->nparams;
      u_int16_t *lens = (u_int16_t *)((char *)ev + sizeof(struct sysdig_event_header));
      char *valptr = (char *)lens + nparams * sizeof(u_int16_t);
      u_int32_t fd = *((u_int32_t*)valptr);
      // u_int32_t size = *((u_int32_t*)(valptr+lens[0]));
      char *tuple  = (char*)((valptr+lens[0]+lens[1]));
      u_int8_t ip_version = tuple[0];
      u_int sock_type;
      u_int16_t pid;
      char *process_name;

      if((ip_version != AF_INET) && (ip_version != AF_INET6)) return;

      readWriteGlobals->sysdig.last_fd[cpuid] = fd;
      pid = tid_to_pid(tid, &process_name);
      sock_type = get_socket_type(pid, readWriteGlobals->sysdig.last_fd[cpuid]);

      if((pid == my_tid) && (sock_type == (u_int)-1 /* Not found */)) {
	/* If is UDP the tuple is NOT NULL (Optimization) */
	u_int16_t m_sport  = *((u_int16_t*)(&tuple[1+4]));      
	
	if(m_sport == 0) sock_type = SOCK_STREAM; else sock_type = SOCK_DGRAM;
      }

      if(sock_type != (u_int)-1 /* Found */) {
	PRINT_TRACE();
	create_flow(ev, i, 1, readWriteGlobals->sysdig.last_fd[cpuid], tuple,
		    (sock_type == SOCK_STREAM) ? IPPROTO_TCP : IPPROTO_UDP,
		    (ip_version == AF_INET) ? 4 : 6);
      }
    }
    break;

  default:
    //traceEvent(TRACE_NORMAL, "Unknown event %u\n", ev->event_type);
    break;
  }

  if(0)
    traceEvent(TRACE_INFO, "%lu)\t[tid=%lu][len=%u][type=%u/%s]: done\n", 
	       ev->ts, ev->thread_id, ev->event_len, ev->event_type, i->name);

  readWriteGlobals->sysdig.last_ts = ev->ts;
}


